@@ -1,9 +1,13 @@
 """lmdierct"""
-from lmdirect.const import DISABLED, ENABLED
+from lmdirect.const import DISABLED, ENABLED, MACHINE_NAME, MODEL_NAME, SERIAL_NUMBER
 from .connection import Connection
 from .msgs import (
+    AUTO_BITFIELD,
     DOSE_TEA,
+    FIRMWARE_VER,
     GLOBAL_AUTO,
+    MON_OFF,
+    MON_ON,
     TYPE_AUTO_ON_OFF,
     TYPE_COFFEE_TEMP,
     TYPE_MAIN,
@@ -14,10 +18,10 @@ from .msgs import (
     AUTO_SCHED_MAP,
     AUTO_BITFIELD_MAP,
     DOSE_K1,
-    TEMP_COFFEE,
-    TEMP_STEAM,
-    TON_PREBREWING_K1,
-    TOFF_PREBREWING_K1,
+    TSET_COFFEE,
+    TSET_STEAM,
+    PREBREWING_TON_K1,
+    PREBREWING_TOFF_K1,
 )
 import asyncio
 
@@ -27,8 +31,8 @@ _LOGGER = logging.getLogger(__name__)
 
 
 class LMDirect(Connection):
-    def __init__(self, creds):
-        super().__init__(creds)
+    def __init__(self, machine_info):
+        super().__init__(machine_info)
 
     @property
     def current_status(self):
@@ -38,17 +42,22 @@ class LMDirect(Connection):
     @property
     def machine_name(self):
         """Return the name of the machine"""
-        return self._machine_name
+        return self._machine_info[MACHINE_NAME]
 
     @property
     def serial_number(self):
         """Return serial number"""
-        return self._serial_number
+        return self._machine_info[SERIAL_NUMBER]
 
     @property
     def model_name(self):
         """Return model name"""
-        return self._model_name
+        return self._machine_info[MODEL_NAME]
+
+    @property
+    def firmware_version(self):
+        """Return firmware version"""
+        return self._current_status.get(FIRMWARE_VER, "Unknown")
 
     def register_callback(self, callback):
         """Register callback for updates"""
@@ -67,15 +76,21 @@ class LMDirect(Connection):
         if key in self._raw_callback_list:
             self._raw_callback_list.remove(key)
 
+    async def connect(self):
+        return await self._connect()
+
     async def request_status(self):
         """Request all status elements"""
+        msgs = [
+            Msg.GET_STATUS,
+            Msg.GET_CONFIG,
+            Msg.GET_AUTO_SCHED,
+            Msg.GET_DATETIME,
+            Msg.GET_DRINK_STATS,
+        ]
 
         _LOGGER.debug("Requesting status")
-        await self._send_msg(Msg.GET_STATUS)
-        await self._send_msg(Msg.GET_CONFIG)
-        await self._send_msg(Msg.GET_AUTO_SCHED)
-        # await self._send_msg(Msg.GET_SER_NUM)
-        await self._send_msg(Msg.GET_DATETIME)
+        await asyncio.gather(*[self._send_msg(msg) for msg in msgs])
 
         """Also wait for current temp"""
         self._responses_waiting.append(MSGS[Msg.GET_TEMP_REPORT].msg)
@@ -86,7 +101,8 @@ class LMDirect(Connection):
 
     async def close(self):
         """Wait for the read task to exit"""
-        await asyncio.gather(*[self._reaper_task])
+        if self._read_reaper_task:
+            await asyncio.gather(self._read_reaper_task)
         await self._close()
 
     def call_callbacks(self, **kwargs):
@@ -111,11 +127,11 @@ class LMDirect(Connection):
     async def set_power(self, power):
         """Send power on or power off commands"""
         value = self.convert_to_ascii(0x01 if power else 0x00, size=1)
-        await self._send_msg(Msg.SET_POWER, value)
+        await self._send_msg(Msg.SET_POWER, data=value)
         await self._send_msg(Msg.GET_CONFIG)
         self._call_callbacks(entity_type=TYPE_MAIN)
 
-    async def set_auto_on_off(self, field_name, value):
+    async def set_auto_on_off(self, day_of_week=None, enable=None):
         """Configure auto on/off"""
 
         async def auto_on_off_callback(key, data):
@@ -127,7 +143,7 @@ class LMDirect(Connection):
 
             try:
                 """Find the bit"""
-                elem = self.findkey("AUTO_BITFIELD", AUTO_SCHED_MAP)
+                elem = self.findkey(AUTO_BITFIELD, AUTO_SCHED_MAP)
 
                 """The strings are ASCII-encoded hex, so each value takes 2 bytes"""
                 index = elem.index * 2
@@ -135,21 +151,27 @@ class LMDirect(Connection):
 
                 """Extract value for this field"""
                 orig_value = int(data[index : index + size], 16)
-                bitmask = 0x01 << self.findkey(field_name, AUTO_BITFIELD_MAP)
+                bitmask = 0x01 << self.findkey(day_of_week, AUTO_BITFIELD_MAP)
                 new_value = self.convert_to_ascii(
-                    orig_value | bitmask if value else orig_value & ~bitmask, 1
+                    orig_value | bitmask if enable else orig_value & ~bitmask, 1
                 )
 
                 buf_to_send = data[:index] + new_value + data[index + size :]
 
-                self._current_status[field_name] = ENABLED if value else DISABLED
+                self._current_status[day_of_week] = ENABLED if enable else DISABLED
 
-                await self._send_msg(Msg.SET_AUTO_SCHED, buf_to_send)
+                await self._send_msg(Msg.SET_AUTO_SCHED, data=buf_to_send)
                 await self._send_msg(Msg.GET_AUTO_SCHED)
+
                 self._call_callbacks(entity_type=TYPE_AUTO_ON_OFF)
             except Exception as err:
                 _LOGGER.debug(f"Caught exception: {err}")
 
+        _LOGGER.error(f"Called with day_of_week:{day_of_week} enable:{enable}")
+        if None in [day_of_week, enable]:
+            msg = f"Some parameters invalid {day_of_week} {enable}"
+            raise InvalidInput(msg)
+
         self.register_raw_callback(Msg.GET_AUTO_SCHED, auto_on_off_callback)
         await self._send_msg(Msg.GET_AUTO_SCHED)
 
@@ -157,7 +179,9 @@ class LMDirect(Connection):
         """Set global auto on/off"""
         await self.set_auto_on_off(GLOBAL_AUTO, value)
 
-    async def set_auto_on_off_hours(self, field_name, hour_on, hour_off):
+    async def set_auto_on_off_hours(
+        self, day_of_week=None, hour_on=None, hour_off=None
+    ):
         """Configure auto on/off"""
 
         async def auto_on_off_callback(key, data):
@@ -169,7 +193,7 @@ class LMDirect(Connection):
 
             try:
                 """Find the "on" element"""
-                elem = self.findkey(field_name.replace("AUTO", "ON"), AUTO_SCHED_MAP)
+                elem = self.findkey(day_of_week + MON_ON[3:], AUTO_SCHED_MAP)
 
                 """The strings are ASCII-encoded hex, so each value takes 2 bytes and there 2 of them"""
                 index = elem.index * 2
@@ -183,23 +207,30 @@ class LMDirect(Connection):
                     + data[index + size :]
                 )
 
-                self._current_status[field_name.replace("AUTO", "ON")] = hour_on
-                self._current_status[field_name.replace("AUTO", "OFF")] = hour_off
+                self._current_status[day_of_week + MON_ON[3:]] = hour_on
+                self._current_status[day_of_week + MON_OFF[3:]] = hour_off
 
-                _LOGGER.info(f"Orig data: {data}")
-                _LOGGER.info(f"New data:  {buf_to_send}")
-                await self._send_msg(Msg.SET_AUTO_SCHED, buf_to_send)
+                await self._send_msg(Msg.SET_AUTO_SCHED, data=buf_to_send)
                 await self._send_msg(Msg.GET_AUTO_SCHED)
                 self._call_callbacks(entity_type=TYPE_AUTO_ON_OFF)
             except Exception as err:
                 _LOGGER.debug(f"Caught exception: {err}")
+                raise
+
+        if None in [day_of_week, hour_on, hour_off]:
+            msg = f"Some parameters invalid {day_of_week} {hour_on} {hour_off}"
+            raise InvalidInput(msg)
 
         self.register_raw_callback(Msg.GET_AUTO_SCHED, auto_on_off_callback)
         await self._send_msg(Msg.GET_AUTO_SCHED)
 
-    async def set_dose(self, key, pulses):
+    async def set_dose(self, key=None, pulses=None):
         """Set the coffee boiler temp in Celcius"""
 
+        if None in [key, pulses]:
+            msg = f"set_dose: Some parameters not specified {key} {pulses}"
+            raise InvalidInput(msg)
+
         if isinstance(pulses, str):
             pulses = int(pulses)
 
@@ -208,8 +239,8 @@ class LMDirect(Connection):
 
         """Validate input"""
         if not (1 <= pulses <= 1000 and 1 <= key <= 5):
-            _LOGGER.error(f"Invalid values pulses:{pulses} key:{key}")
-            return False
+            msg = f"set_dose: Invalid values pulses:{pulses} key:{key}"
+            raise InvalidInput(msg)
 
         self._current_status[DOSE_K1.replace("1", str(key))] = pulses
 
@@ -219,16 +250,19 @@ class LMDirect(Connection):
         await self._send_msg(Msg.GET_CONFIG)
         self._call_callbacks(entity_type=TYPE_MAIN)
 
-    async def set_dose_tea(self, seconds):
+    async def set_dose_tea(self, seconds=None):
         """Set the coffee boiler temp in Celcius"""
 
+        if not seconds:
+            raise InvalidInput("set_dose_tea: Seconds not specified")
+
         if isinstance(seconds, str):
             seconds = int(seconds)
 
         """Validate input"""
         if not (1 <= seconds <= 30):
-            _LOGGER.error(f"Invalid values seconds:{seconds}")
-            return False
+            msg = f"Invalid values seconds:{seconds}"
+            raise InvalidInput(msg)
 
         self._current_status[DOSE_TEA] = seconds
 
@@ -237,64 +271,73 @@ class LMDirect(Connection):
         await self._send_msg(Msg.GET_CONFIG)
         self._call_callbacks(entity_type=TYPE_MAIN)
 
-    async def set_prebrew_times(self, key, time_on, time_off):
+    async def set_prebrew_times(self, key=None, time_on=None, time_off=None):
         """Set the coffee boiler temp in Celcius"""
 
+        if None in [key, time_on, time_off]:
+            raise InvalidInput(
+                "set_prebrew_times: Some parameters invalid {key} {time_on} {time_off}"
+            )
+
         if isinstance(time_on, str):
             time_on = float(time_on)
 
         if isinstance(time_off, str):
-            off_time = float(time_off)
+            time_off = float(time_off)
 
         if isinstance(key, str):
             key = int(key)
 
         """Validate input"""
         if not (0 <= time_on <= 5.9 and 0 <= time_off <= 5.9 and 1 <= key <= 4):
-            _LOGGER.error(
-                f"Invalid values time_on:{time_on} off_time:{time_off} key:{key}"
-            )
-            return False
+            msg = f"Invalid values time_on:{time_on} off_time:{time_off} key:{key}"
+            raise InvalidInput(msg)
 
-        self._current_status[TON_PREBREWING_K1.replace("1", str(key))] = time_on
-        self._current_status[TOFF_PREBREWING_K1.replace("1", str(key))] = time_off
+        self._current_status[PREBREWING_TON_K1.replace("1", str(key))] = time_on
+        self._current_status[PREBREWING_TOFF_K1.replace("1", str(key))] = time_off
 
         """set "on" time"""
         key_on = self.convert_to_ascii(Msg.PREBREW_ON_BASE + (key - 1), size=1)
         data = self.convert_to_ascii(int(time_on * 10), size=1)
-        await self._send_msg(Msg.SET_PB_ON, key=key_on, data=data)
+        await self._send_msg(Msg.SET_PREBREW_TIMES, key=key_on, data=data)
 
         """set "off" time"""
         key_off = self.convert_to_ascii(Msg.PREBREW_OFF_BASE + (key - 1), size=1)
         data = self.convert_to_ascii(int(time_off * 10), size=1)
-        await self._send_msg(Msg.SET_PB_OFF, key=key_off, data=data)
+        await self._send_msg(Msg.SET_PREBREW_TIMES, key=key_off, data=data)
         await self._send_msg(Msg.GET_CONFIG)
         self._call_callbacks(entity_type=TYPE_PREBREW)
 
-    async def set_coffee_temp(self, temp):
+    async def set_coffee_temp(self, temp=None):
         """Set the coffee boiler temp in Celcius"""
 
+        if not temp:
+            raise InvalidInput("set_coffee__temp: Temperature not specified")
+
         if isinstance(temp, str):
             temp = float(temp)
 
         temp = round(temp, 1)
 
-        self._current_status[TEMP_COFFEE] = temp
+        self._current_status[TSET_COFFEE] = temp
 
         data = self.convert_to_ascii(int(temp * 10), size=2)
         await self._send_msg(Msg.SET_COFFEE_TEMP, data=data)
         await self._send_msg(Msg.GET_TEMP_REPORT, data=data)
         self._call_callbacks(entity_type=TYPE_COFFEE_TEMP)
 
-    async def set_steam_temp(self, temp):
+    async def set_steam_temp(self, temp=None):
         """Set the steam boiler temp in Celcius"""
 
+        if not temp:
+            raise InvalidInput("set_steam_temp: Temperature not specified")
+
         if isinstance(temp, str):
             temp = float(temp)
 
         temp = round(temp, 1)
 
-        self._current_status[TEMP_STEAM] = temp
+        self._current_status[TSET_STEAM] = temp
 
         data = self.convert_to_ascii(int(temp * 10), size=2)
         await self._send_msg(Msg.SET_STEAM_TEMP, data=data)
@@ -306,4 +349,11 @@ class LMDirect(Connection):
         data = self.convert_to_ascii(0x01 if enable else 0x00, size=1)
         await self._send_msg(Msg.SET_PREBREWING_ENABLE, data=data)
         await self._send_msg(Msg.GET_CONFIG)
-        self._call_callbacks(entity_type=TYPE_PREBREW)
\ No newline at end of file
+        self._call_callbacks(entity_type=TYPE_PREBREW)
+
+
+class InvalidInput(BaseException):
+    """Error to indicate there is no Connection"""
+
+    def __init__(msg):
+        _LOGGER.error(msg)
\ No newline at end of file
